"""
Defines the current individuals in the population.
Responsible for generating the next generation
"""
from galapagos.utils import Utils
from typing import Self
from collections import defaultdict


class Generation:
    def __init__(self, population: list = [], genepool: list = []):
        self.__genepool = genepool
        self.population = population 

    def __len__(self):
        partial_size = 0
        for item in self.population:
            individual = item[0]
            individual_count = item[1]
            partial_size += individual_count
        return partial_size

    def __getitem__(self, key):
        return self.population[key]

    def __setitem__(self, key, value):
        self.population[key] = value

    def __str__(self):
        return_string = ""
        for i in range(len(self.population)):
            return_string += f"Indivíduo: {self[i][0]} Quantidade: {self[i][1]} \n"
        return return_string

    @property
    def genepool(self):
        return self.__genepool

    def next(self) -> Self:
        next_gen = Generation(genepool=self.genepool)
        male_available_individuals = []
        female_available_individuals = []
        for item in self.population:
            individual = item[0]
            if individual.sex == "M":
                male_available_individuals.append(item)
            else:
                female_available_individuals.append(item)

        next_gen_individual_counts = defaultdict(int)
        for i in range(len(self)):
            father = Utils.select_individual(male_available_individuals)
            mother = Utils.select_individual(female_available_individuals)
            newborn = father.mate(mother)
            newborn.update_fitness(self.genepool)

            next_gen_individual_counts[newborn] += 1

        next_gen.population = [
            [individual, individual_count] for individual, individual_count in next_gen_individual_counts.items()
        ]

        return next_gen




import random


class Utils:
    def random_sex() -> str:
        return "M" if random.randint(0, 1) == 0 else "F"

    def get_total_cumulative_fit(available_individuals: list):
        cumulative_fit = 0
        for individual in available_individuals:
            cumulative_fit += individual[0].fitness * individual[1]
        return cumulative_fit

    def select_individual(available_individuals: list) -> 'Individual':
        # In probability theory, this is known
        # as sampling from a discrete distribution.
        cumulative_fit = 0
        total_cumulative_fit = (Utils
                                .get_total_cumulative_fit(available_individuals)
                                )
        rand = random.uniform(0, total_cumulative_fit)
        for individual in available_individuals:
            cumulative_fit += individual[0].fitness * individual[1]
            if cumulative_fit > rand:
                return individual[0]

        if cumulative_fit <= rand:
            raise Exception("Valor aleatório não mapeia para "
                            "nenhum indivíduo")



"""
Defines an individual to simulate.
"""
from galapagos.genotype import Genotype
from galapagos.locus import Locus
from galapagos.utils import Utils
from typing import Self


class Individual:
    def __init__(self, sex: str, genotype: Genotype,
                 fitness: float = 1.0):
        self.__fitness = fitness
        self.__sex = sex
        self.genotype = genotype

    def __str__(self):
        return (f"Sexo: {self.sex} Genótipo: {self.genotype}"
                f"Fitness: {self.__fitness}")

    def __eq__(self, other):
        return (self.fitness == other.fitness) and (self.sex == other.sex) and (self.genotype == other.genotype)

    def __hash__(self):
        return hash((self.fitness, self.sex, self.genotype))

    @property
    def fitness(self):
        return self.__fitness

    @fitness.setter
    def fitness(self, fitness):
        self.__fitness = fitness

    @property
    def sex(self):
        return self.__sex

    def mate(self, other: Self) -> Self:
        new_genotype = Genotype(size=len(self.genotype))
        for i in range(len(self.genotype)):
            locus = Locus( (self.genotype[i].pass_allele(),
                            other.genotype[i].pass_allele())
                          )
            new_genotype[i] = locus

        new_sex = Utils.random_sex()
        newborn = Individual(new_sex, new_genotype)

        return newborn

    def update_fitness(self, genepool):
        if genepool:
            self.fitness = 0.0
            for gene in genepool:
                if gene[0] in self.genotype:
                    self.fitness += gene[1]/len(self.genotype)



from galapagos.generation import Generation
from galapagos.utils import Utils
from galapagos.genotype import Genotype
from galapagos.individual import Individual

class Genepool:
    def __init__(self, 
                 genotypes: list, 
                 fitnesses: list,
                 loci: list
                 ):
        self.genotypes = genotypes
        self.fitnesses = fitnesses
        self.loci = loci

    def __getitem__(self, key):
        return (self.genotype[key], self.fitnesses[key], self.loci[key])

    def __setitem__(self, key, value):
        self.genotype[key] = value[0]
        self.fitnesses[key] = value[1]
        self.loci[key] = value[2]





"""
Defines a given genotype
"""


class Genotype:
    def __init__(self, size: int):
        self.__size = size
        self.__loci = [None] * size  # Gera uma genótipo vazio com
                                     # `size` loci

    def __str__(self):
        return_string = ""
        for i in range(len(self)):
            return_string += f"{self[i]} "
        return return_string

    def __len__(self):
        return self.__size

    def __getitem__(self, key):
        return self.__loci[key]

    def __setitem__(self, key, value):
        self.__loci[key] = value

    def __contains__(self, locus):
        for item in self:
            if item == locus:
                return True
        return False

    def __eq__(self, other):
        if len(self) != len(other):
            return False
        for i in range(len(self)):
            if self[i] != other[i]:
                return False

        return True

    def __hash__(self):
        return hash(tuple(self.__loci))

    @property
    def loci(self):
        return self.__loci



import random
from typing import Self


class Locus:
    def __init__(self, alleles: tuple):
        self.alleles = alleles  

    def __str__(self):
        canon_alleles = sorted(self.alleles)
        return_str = f""
        for i in canon_alleles:
            return_str += f"{i}"
        return return_str

    def __eq__(self, other: Self):
        return set(self.alleles) == set(other.alleles)

    def __hash__(self):
        return hash(frozenset(self.alleles))

    def __getitem__(self, key):
        return sorted(self.alleles)[key]

    def __len__(self):
        return len(self.alleles)

    def pass_allele(self):
        rand = random.randint(0, len(self.alleles) - 1)
        return self.alleles[rand]



from galapagos.generation import Generation
from galapagos.utils import Utils
from galapagos.genotype import Genotype
from galapagos.individual import Individual
from plotnine import *
from collections import defaultdict
import plotly.express as px
import numpy as np
import polars as pl
import math

class Simulation:
    def __init__(self,
                 max_generations: int,
                 first_generation: Generation,
                 genepool: list
                 ):
        self.max_generations = max_generations
        self.genepool = genepool
        self.data = []

        self.first_generation = first_generation
        self.generation_history = [self.first_generation]

    # TODO: Still under development
    def _get_individual_allele_frequency(self, 
                                         freq_dict,
                                         individual: Individual, 
                                         individual_count: str):
        for locus in individual.genotype:
            for allele in locus:
                freq_dict[allele] += individual_count

    def _process_generation_data(self, generation, index: int):
        freq_dict = defaultdict(int)
        for individual, individual_count in generation:
            self._get_individual_allele_frequency(freq_dict, individual, individual_count)

        locus = generation[0][0].genotype.loci[0]
        for allele, allele_count in freq_dict.items():
            self.data.append({
                "generation": index,
                "allele": str(allele),
                "frequency": allele_count / (len(locus) * len(generation))
            })

    def _selection_model(self) -> list:
        # TODO: Rever referência, me parece errado
        data_theory = []
        first_gen = self.first_generation
        freq_prev = {}
        freq_theory = {}

        for locus, _ in self.genepool:
            freq_prev[str(locus)] = first_gen.get_locus_frequency(locus)

        for g in range(self.max_generations):
            fitness_avg = 0

            # Calculates fitness_avg
            for locus, fitness in self.genepool:
                fitness_avg += freq_prev[str(locus)] * fitness

            # Calculate expected frequency
            for locus, fitness in self.genepool:
                freq_theory[str(locus)] = freq_prev[str(locus)] * fitness / fitness_avg
                data_theory.append({
                    "generation": g + 1,
                    "genotype": str(locus),
                    "expected_frequency": freq_theory[str(locus)]
                })
            freq_prev = freq_theory
        return data_theory

    def get_allele_frequency(self, allele: str):
        df = pl.DataFrame(self.data)
        return df.filter(pl.col("allele") == allele).sum()

    def simulate(self):
        gene_lines = [
            f"Gene: {str(gene)}, Fitness: {fitness}"
            for gene, fitness in self.genepool
        ]

        print(f"Simulation info: \n"
              f"maximum number of generations: {self.max_generations}, \n"
              f"population size: {len(self.first_generation)}, \n"
              f"available genes: {gene_lines}"
             )
        for i in range(self.max_generations):
            new_generation = self.generation_history[i].next()
            self.generation_history.append(new_generation)
            self._process_generation_data(new_generation, index=i + 1)

    def show(self):

        df_simu = pl.DataFrame(self.data)
        # fig = px.line(df_simu,
        #               x="generation",
        #               y="frequency",
        #               color="allele",
        #               title="Simulação de Frequência de Genótipo")

        # fig.show()
        # df_theory = pl.DataFrame(self._selection_model())
        (
            ggplot(df_simu, aes(x="generation", color="allele"))
            + geom_line(aes(y="frequency"), size=1)
            + scale_y_continuous(limits=(0, 1))
            # + geom_line(data=df_theory,
            #            mapping=aes(y="expected_frequency", linetype="'Teórico'"),
            #            size=1
            #            )
            + labs(x="Geração", y="Frequência", color="Alelo")

            # + scale_linetype_manual(
            #     name="Fonte dos Dados",
            #     values={"Simulado": "solid", "Teórico": "dashed"}
            # )
        ).show()
